# -*- coding: utf-8 -*-

"""PyKEEN is a package for training and evaluating knowledge graph embedding.

Installation
------------
Get the Latest
~~~~~~~~~~~~~~~
Download the most recent code from ... with:

For Developers
~~~~~~~~~~~~~~
Clone the repository from `GitHub <https://github.com/SmartDataAnalytics/pykeen.git>'
"""

from .constants import get_version  # noqa: F401
from .run import Results, run  # noqa: F401
